import os
from typing import Any
import requests
import json

from ..load_file import load_file
from ..store_file_local import store_file_local, store_json_local


class ZenodoUploader:
    def __init__(self, *,
        author: str,
        affiliation: str,
        title: str,
        description: str,
        sandbox: bool=True
    ) -> None:
        self._author = author
        self._affiliation = affiliation
        self._title = title
        self._description = description
        self._sandbox = sandbox
        self._base_zenodo_url = 'https://sandbox.zenodo.org' if sandbox else 'https://zenodo.org'
        env_name = 'ZENODO_SANDBOX_ACCESS_TOKEN' if sandbox else 'ZENODO_ACCESS_TOKEN'
        try:
            self._access_token = os.environ[env_name]
        except:
            raise Exception(f'Environment variable not set: {env_name}')
        
        # create a new deposition
        print('Creating new deposition...')
        headers = {"Content-Type": "application/json"}
        r2 = requests.post(
            f'{self._base_zenodo_url}/api/deposit/depositions',
            params={'access_token': self._access_token},
            json={},
            headers=headers
        )
        if r2.status_code not in [200, 201]:
            raise Exception(f'Problem creating deposition: {r2.status_code}')

        # get the deposition ID
        self._deposition_id = r2.json()['id']
        print(f'Deposition ID: {self._deposition_id}')

        # get the bucket url
        self._bucket_url = r2.json()["links"]["bucket"]
        print(f'Bucket URL: {self._bucket_url}')
    def upload_file(self, uri: str, *, name: str):
        path = load_file(uri)
        print(f'Uploading {name}...')
        with open(path, 'rb') as ff:
            r3 = requests.put(
                f'{self._bucket_url}/{name}',
                params={'access_token': self._access_token},
                data=ff
            )
            if r3.status_code not in [200, 201]:
                raise Exception(f'Problem uploading file: {r3.status_code}')
        return f'{"zenodo-sandbox" if self._sandbox else "zenodo"}://{self._deposition_id}/{name}'
    def upload_file_recursive(self, uri: str, *, name: str):
        print(f'Uploading recursive {name}: {uri}')
        file_path = load_file(uri)
        if file_path is None:
            raise Exception(f'Unable to load {uri}')
        try:
            with open(file_path, 'r') as ff:
                file_obj = json.load(ff)
        except:
            file_obj = None
            print(f'Warning: unable to load json from file: {uri}')
        if file_obj is not None:
            file_obj_processed = self._upload_file_recursive_process_obj(file_obj)
            uri2 = store_json_local(file_obj_processed)
            return self.upload_file(uri2, name=name)
        else:
            return self.upload_file(uri, name=name)
    def _upload_file_recursive_process_obj(self, x: Any):
        if isinstance(x, dict):
            ret = {}
            for k, v in x.items():
                ret[k] = self._upload_file_recursive_process_obj(v)
            return ret
        elif isinstance(x, list):
            ret = []
            for i in range(len(x)):
                ret.append(self._upload_file_recursive_process_obj(x[i]))
            return ret
        elif isinstance(x, str):
            if x.startswith('sha1://') or x.startswith('zenodo://') or x.startswith('zenodo-sandbox://'):
                uri0 = store_file_local(x) # make sure it's a sha-1 url
                sha1 = uri0.split('?')[0].split('/')[2]
                return self.upload_file_recursive(x, name=sha1)
            else:
                return x
        else:
            return x
    def finalize_upload(self):
        data = {
            'metadata': {
                'title': self._title,
                'upload_type': 'dataset',
                'description': self._description,
                'creators': [{'name': self._author, 'affiliation': self._affiliation}]
            }
        }
        # put the meta data
        print('Uploading metadata')
        headers = {"Content-Type": "application/json"}
        r4 = requests.put(
            f'{self._base_zenodo_url}/api/deposit/depositions/{self._deposition_id}',
            params={'access_token': self._access_token},
            data=json.dumps(data),
            headers=headers
        )
        if r4.status_code not in [200, 201]:
            raise Exception(f'Problem uploading metadata: {r4.status_code}')
        rr = r4.json()
        links = rr['links']
        latest_draft_html = links['latest_draft_html']
        print('')
        print(f'IMPORTANT: You must visit this link to finalize the upload:')
        print(f'{latest_draft_html}')
        input('ZENODO UPLOAD: Once you have visited the above link and published this zenodo upload, press enter to continue. ')

def initiate_zenodo_upload(*,
    author: str,
    affiliation: str,
    title: str,
    description: str,
    sandbox: bool=True
):  
    uploader = ZenodoUploader(
        author=author,
        affiliation=affiliation,
        title=title,
        description=description,
        sandbox=sandbox
    )
    return uploader
    




import json
from multiprocessing.dummy import shutdown
import os
import shutil
from typing import Union
import requests
import random

from kachery_cloud.mutable_local import get_mutable_local, set_mutable_local
from .get_kachery_cloud_dir import get_kachery_cloud_dir
from ._kacherycloud_request import _kacherycloud_request
from .store_file_local import _compute_file_hash
from ._fs_operations import _makedirs, _chmod_file
from ._access_group_encrypt import _access_group_decrypt
from .encrypt_uri import decrypt_uri
from ._load_github_file import _load_github_file


def load_file(uri: str, *, verbose: bool=False, local_only: bool=False, dest: Union[None, str]=None, _get_info: bool=False) -> Union[str, dict, None]:
    if uri.startswith('sha1-enc://'):
        # While it would be nice to send the sha1-enc hash directly to the API findFile,
        # we need to do it this way so that we can check locally... because files are stored
        # locally by the sha1 hash. Unfortunately, this means two http calls.
        uri_decrypted = decrypt_uri(uri)
        return load_file(uri_decrypted, verbose=verbose, local_only=local_only, dest=dest, _get_info=_get_info)
    
    if uri.startswith('jot://'):
        jot_id = uri.split('?')[0].split('/')[2]
        uri2 = _get_jot_value(jot_id)
        if uri2 is None:
            raise Exception('Unable to get jot')
        uri = uri2
    
    if uri.startswith('github://'):
        return _load_github_file(uri)

    if local_only:
        return load_file_local(uri, dest=dest)
    if uri.startswith('/'):
        if _get_info:
            raise Exception('Cannot use _get_info for this uri')
        if os.path.exists(uri):
            if dest is not None:
                shutil.copyfile(uri, dest)
                return dest
            return uri
        else:
            return None
    if uri.startswith('sha1://'):
        if not _get_info:
            x = load_file_local(uri, dest=dest)
            if x is not None:
                return x
        sha1 = uri.split('?')[0].split('/')[2]
        return _load_sha1_file_from_cloud(sha1, verbose=verbose, dest=dest, _get_info=_get_info)
    
    if uri.startswith('zenodo://') or uri.startswith('zenodo-sandbox://'):
        x = load_file_local(uri, dest=dest)
        if x is not None:
            return x
        return _load_zenodo_file_from_cloud(uri, dest=dest)

    if _get_info:
        raise Exception('Cannot use _get_info for this uri')

    assert uri.startswith('ipfs://'), f'Invalid or unsupported URI: {uri}'
    a = uri.split('?')[0].split('/')
    assert len(a) >= 3, f'Invalid or unsupported URI: {uri}'
    cid = a[2]

    kachery_cloud_dir = get_kachery_cloud_dir()
    e = cid[-6:]
    parent_dir = f'{kachery_cloud_dir}/ipfs/{e[0]}{e[1]}/{e[2]}{e[3]}/{e[4]}{e[5]}'
    filename = f'{parent_dir}/{cid}'
    if os.path.exists(filename):
        if dest is not None:
            shutil.copyfile(filename, dest)
            return dest
        return filename

    payload = {
        'type': 'findIpfsFile',
        'cid': cid
    }
    response= _kacherycloud_request(payload)
    found = response['found']
    if found:
        url = response['url']
    else:
        return None
        # url = f'https://{cid}.ipfs.dweb.link'
        # url = f'https://cloudflare-ipfs.com/ipfs/{cid}'
        # url = f'https://ipfs.filebase.io/ipfs/{cid}'

    if verbose:
        print(f'Loading file from kachery cloud: {uri}')    
    if not os.path.exists(parent_dir):
        _makedirs(parent_dir)
    tmp_filename = f'{filename}.tmp.{_random_string(8)}'
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(tmp_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    try:
        os.rename(tmp_filename, filename)
        # _chmod_file(filename)
    except:
        if not os.path.exists(filename): # maybe some other process beat us to it
            raise Exception(f'Unexpected problem moving file {tmp_filename}')
    if dest is not None:
        shutil.copyfile(filename, dest)
        return dest
    return filename

def load_file_info(uri: str) -> dict:
    return load_file(uri, _get_info=True)

def _load_zenodo_file_from_cloud(uri: str, *, dest: Union[None, str]=None):
    sandbox = uri.startswith('zenodo-sandbox://')
    aa = uri.split('?')[0].split('/')
    z_record_id = aa[2]
    z_file_name = '/'.join(aa[3:])

    record_url = f'https://{"sandbox.zenodo" if sandbox else "zenodo"}.org/api/records/{z_record_id}'
    req_resp = requests.get(record_url)
    if req_resp.status_code != 200:
        raise Exception(f'Error downloading zenodo record info ({req_resp.status_code}) {req_resp.reason}: {req_resp.text}')
    resp = req_resp.json()
    if 'files' not in resp:
        raise Exception('Error getting zenodo record')
    cc = [f for f in resp['files'] if f.get('filename', f.get('key', '')) == z_file_name]
    if len(cc) == 0:
        raise Exception('File not found in zenodo record')
    dd = cc[0]
    url = dd['links'].get('download', dd['links'].get('self'))
    
    kachery_cloud_dir = get_kachery_cloud_dir()
    filename = f'{kachery_cloud_dir}/{"zenodo-sandbox" if sandbox else "zenodo"}/{z_record_id}/{z_file_name}'
    parent_dir = os.path.dirname(filename)
    if not os.path.exists(parent_dir):
        os.makedirs(parent_dir)
    tmp_filename = f'{filename}.tmp.{_random_string(8)}'
    with requests.get(url, stream=True) as r:
        if r.status_code == 404:
            raise Exception(f'Unexpected: file not found: {url}')
        r.raise_for_status()
        with open(tmp_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    try:
        os.rename(tmp_filename, filename)
        # _chmod_file(filename)
    except:
        if not os.path.exists(filename): # maybe some other process beat us to it
            raise Exception(f'Unexpected problem moving file {tmp_filename}')
    if dest is not None:
        shutil.copyfile(filename, dest)
        return dest
    return filename

def _load_sha1_file_from_cloud(sha1: str, *, verbose: bool, dest: Union[None, str]=None, _get_info: bool=False) -> Union[str, dict, None]:
    payload = {
        'type': 'findFile',
        'hashAlg': 'sha1',
        'hash': sha1
    }
    if os.environ.get('USE_KACHERY_GATEWAY', '') == '0':
        response= _kacherycloud_request(payload)
    else:
        from ._kachery_gateway_request import _kachery_gateway_request
        response= _kachery_gateway_request(payload)
        
    found = response['found']
    uri = f'sha1://{sha1}'
    if found:
        url = response['url']
    else:
        return None

    if _get_info:
        # we don't want the user to get the idea they should use the URL directly!
        del response['url']
        return response

    kachery_cloud_dir = get_kachery_cloud_dir()
    e = sha1
    parent_dir = f'{kachery_cloud_dir}/sha1/{e[0]}{e[1]}/{e[2]}{e[3]}/{e[4]}{e[5]}'
    filename = f'{parent_dir}/{sha1}'
    if verbose:
        print(f'Loading file from kachery cloud: {uri}') 
    if not os.path.exists(parent_dir):
        _makedirs(parent_dir)
    tmp_filename = f'{filename}.tmp.{_random_string(8)}'
    with requests.get(url, stream=True) as r:
        if r.status_code == 404:
            raise Exception(f'Unexpected: file not found in bucket: {url}')
        r.raise_for_status()
        with open(tmp_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    try:
        os.rename(tmp_filename, filename)
        # _chmod_file(filename)
    except:
        if not os.path.exists(filename): # maybe some other process beat us to it
            raise Exception(f'Unexpected problem moving file {tmp_filename}')
    if dest is not None:
        shutil.copyfile(filename, dest)
        return dest
    return filename

def load_file_local(uri: str, *, dest: Union[None, str]=None) -> Union[str, None]:
    if uri.startswith('zenodo://') or uri.startswith('zenodo-sandbox://'):
        sandbox = uri.startswith('zenodo-sandbox://')
        aa = uri.split('?')[0].split('/')
        z_record_id = aa[2]
        z_file_name = '/'.join(aa[3:])

        kcdir = get_kachery_cloud_dir()
        pp = f'{kcdir}/{"zenodo-sandbox" if sandbox else "zenodo"}/{z_record_id}/{z_file_name}'
        if os.path.exists(pp):
            if dest is not None:
                shutil.copyfile(pp, dest)
                return dest
            return pp
        else:
            return None

    query = _get_query_from_uri(uri)
    assert uri.startswith('sha1://'), f'Invalid local URI: {uri}'
    a = uri.split('?')[0].split('/')
    assert len(a) >= 3, f'Invalid or unsupported URI: {uri}'
    sha1 = a[2]

    kachery_cloud_dir = get_kachery_cloud_dir()

    s = sha1
    parent_dir = f'{kachery_cloud_dir}/sha1/{s[0]}{s[1]}/{s[2]}{s[3]}/{s[4]}{s[5]}'
    filename = f'{parent_dir}/{sha1}'
    if os.path.exists(filename):
        if dest is not None:
            shutil.copyfile(filename, dest)
            return dest
        return filename
    
    if 'location' in query:
        location = query['location']
        if os.path.isabs(location) and os.path.exists(location):
            sha1_2 = _compute_file_hash(location, 'sha1')
            if sha1_2 == sha1:
                if dest is not None:
                    shutil.copyfile(location, dest)
                    return dest
                return location
    
    # check for linked file
    a_txt = get_mutable_local(f'@linked_files/@sha1/{sha1}')
    if a_txt is not None:
        a = json.loads(a_txt)
        path0 = a['path']
        size0 = a['size']
        mtime0 = a['mtime']
        if os.path.exists(path0):
            if os.path.getsize(path0) == size0 and os.stat(path0).st_mtime == mtime0:
                if dest is not None:
                    shutil.copyfile(path0, dest)
                    return dest
                return path0
            sha1_0 = _compute_file_hash(path0, algorithm='sha1')
            if sha1_0 == sha1:
                set_mutable_local(f'@linked_files/@sha1/{sha1}', json.dumps({
                    'path': path0,
                    'size': os.path.getsize(path0),
                    'mtime': os.stat(path0).st_mtime,
                    'sha1': sha1
                }))
                if dest is not None:
                    shutil.copyfile(path0, dest)
                    return dest
                return path0
            else:
                print(f'Warning: sha1 of linked file has changed: {path0} {uri}')
    
    return None

def _get_jot_value(jot_id):
    jot_url = 'https://jot.figurl.org/api/jot'
    req = {
        'type': 'getJotValue',
        'jotId': jot_id
    }
    resp = requests.post(jot_url, json=req)
    if resp.status_code != 200:
        raise Exception(f'Error getting jot value: ({resp.status_code}) {resp.reason}: {resp.text}')
    response = resp.json()
    if response['type'] != 'getJotValue':
        raise Exception('Unexpected problem getting jot value')
    return response['value']

def _get_query_from_uri(uri: str):
    a = uri.split('?')
    ret = {}
    if len(a) < 2: return ret
    b = a[1].split('&')
    for c in b:
        d = c.split('=')
        if len(d) == 2:
            ret[d[0]] = d[1]
    return ret

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))


import os
import random
import atexit
import shutil

def use_sandbox(use: bool=True):
    if use:
        os.environ['KACHERY_CLOUD_USE_SANDBOX'] = '1'
        if not os.path.exists(os.environ['KACHERY_CLOUD_SANDBOX_DIR']):
            # only create the dir when we are actually switching to sandbox mode
            os.makedirs(os.environ['KACHERY_CLOUD_SANDBOX_DIR'])
            # important to do it this way so we don't clean up the temp dir in a subprocesss
            _global['sandbox_dir_to_cleanup'] = os.environ['KACHERY_CLOUD_SANDBOX_DIR']
    else:
        # We turn off the sandbox, but if we turn it back on again, we want to use the same dir for this process
        os.environ['KACHERY_CLOUD_USE_SANDBOX'] = '0'

def get_kachery_cloud_dir(*, respect_sandbox: bool=True):
    from pathlib import Path
    if respect_sandbox and (os.getenv('KACHERY_CLOUD_USE_SANDBOX', '') == '1'):
        return os.environ['KACHERY_CLOUD_SANDBOX_DIR']
    homedir = str(Path.home())
    hsd = os.getenv('KACHERY_CLOUD_DIR', f'{homedir}/.kachery-cloud')
    if not os.path.exists(hsd):
        os.makedirs(hsd)
    return hsd

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))

# We use environ variables for this so that the info is passed on to subprocesses
if os.getenv('KACHERY_CLOUD_SANDBOX_DIR', None) is None:
    # We don't actually mkdir until it is used
    # I found that electron remaps the /tmp directory, so we can't use /tmp. Probably just as well.
    os.environ['KACHERY_CLOUD_SANDBOX_DIR'] = f'{get_kachery_cloud_dir()}/sandbox/{_random_string(8)}'
if os.getenv('KACHERY_CLOUD_USE_SANDBOX', None) is None:
    os.environ['KACHERY_CLOUD_USE_SANDBOX'] = '0'

_global = {'sandbox_dir_to_cleanup': None}

def at_exit():
    # important to do it this way so we don't clean up the temp dir in a subprocesss
    sandbox_dir_to_cleanup = _global['sandbox_dir_to_cleanup']
    if sandbox_dir_to_cleanup is not None:
        if os.path.exists(sandbox_dir_to_cleanup):
            print(f'Cleaning up sandbox dir: {sandbox_dir_to_cleanup}')
            shutil.rmtree(sandbox_dir_to_cleanup)

atexit.register(at_exit)


import requests
from datetime import datetime
from .get_client_id import get_client_id
from ._client_keys import _sign_message_as_client, _ephemeral_mode
from ._kachery_cloud_api_url import _kachery_cloud_api_url

def _kacherycloud_request(request_payload: dict):
    client_id = get_client_id()
    url = f'{_kachery_cloud_api_url}/api/kacherycloud'
    timestamp = int(datetime.timestamp(datetime.now()) * 1000)
    payload = {**request_payload, **{'timestamp': timestamp}}
    if not _ephemeral_mode():
        req = {
            'payload': payload,
            'fromClientId': client_id,
            'signature': _sign_message_as_client(payload)
        }
    else:
        # ephemeral mode
        if payload['type'] not in ['getProjectBucketBaseUrl', 'findIpfsFile', 'getMutable', 'subscribeToPubsubChannel', 'publishToPubsubChannel', 'getFeedInfo', 'getFeedMessages']:
            raise Exception(f'Unable to perform operation in ephemeral mode: {payload["type"]}')
        req = {
            'payload': payload
        }
    resp = requests.post(url, json=req)
    if resp.status_code != 200:
        raise Exception(f'Error in {payload["type"]} ({resp.status_code}) {resp.reason}: {resp.text}')
    response = resp.json()
    return response


from typing import Union
import os
from kachery_cloud.TaskBackend import TaskBackend
import kachery_cloud as kcl


# This backend service
# will listen for requests on remote computers to upload files
# that are stored locally.
#
# You can store some content locally via
# uri = kcl.store_text_local('random-text-00001')
# Now this file is stored locally but not in the cloud
# On a remote computer you can send a request to upload
# the file. See the file_share_remote_test.py example
# and paste in the desired url and the project ID
# for this backend.

# uri is obtained from kcl.store_*_local(fname) on this computer
def _kachery_store_shared_file(*, uri: str):
    """uploads the shared file when requested. Note that the uri must be sent by the task on the client
    Parameters
    ----------
    uri : str
        the uri for the file
    Raises
    ------
    Exception
        Raises Unable to load file if file can't be loaded
    """
    # impose restrictions on uri here
    if uri != '' and uri is not None:
        if uri.startswith('sha1://'):
            fname = kcl.load_file(uri, local_only=True) # requires kachery-cloud >= 0.1.19
            if fname is not None:
                print(f'storing {fname} in cloud')
                kcl.store_file(fname)
            else:
                raise Exception(f'Unable to load file: {uri}')
        elif uri.startswith('sha1-enc://'):
            # Important not to share files from encrypted URIs
            # This is how we restrict access
            raise Exception('Cannot share file from encrypted URI')
        else:
            raise Exception(f'Unexpected uri: {uri}')
    else:
        raise Exception(f'Invalid uri: {uri}')

def _start_backend(*, project_id: Union[str, None]=None):
    X = TaskBackend(project_id=project_id)
    X.register_task_handler(
        task_type='action',
        task_name='kachery_store_shared_file.1',
        task_function=_kachery_store_shared_file
    )
    print('Sharing local files')

    # Backend will listen for requests to upload a file to kachery cloud
    X.run()

def share_local_files_experimental(*, project_id: Union[str, None]=None):
    _start_backend(project_id=project_id)


import os
import shutil
from typing import Union
from .get_kachery_cloud_dir import get_kachery_cloud_dir
from ._fs_operations import _makedirs, _chmod_file


def set_mutable_local(key: str, value: str):
    _assert_valid_key(key)
    kachery_cloud_dir = get_kachery_cloud_dir()
    file_name = f'{kachery_cloud_dir}/mutables/{key}'
    parent_dir = os.path.dirname(file_name)
    if not os.path.exists(parent_dir):
        _makedirs(parent_dir)
    with open(file_name, 'w') as f:
        f.write(value)
    # _chmod_file(file_name)

def _assert_valid_key(key: str):
    a = key.split('/')
    if len(a) < 1: raise Exception(f'Invalid mutable key: {key}')
    for i in range(len(a)):
        b = a[i]
        if i < len(a) - 1:
            if not b.startswith('@'):
                raise Exception(f'Invalid mutable key: {key}')
        else:
            if b.startswith('@'):
                raise Exception(f'Invalid mutable key: {key}')
        if b == '.' or b == '..':
            raise Exception(f'Invalid mutable key: {key}')

def _assert_valid_folder_key(key: str):
    a = key.split('/')
    if len(a) < 1: raise Exception(f'Invalid mutable folder key: {key}')
    for i in range(len(a)):
        b = a[i]
        if not b.startswith('@'):
            raise Exception(f'Invalid mutable folder key: {key}')
        if b == '.' or b == '..':
            raise Exception(f'Invalid mutable folder key: {key}')

def get_mutable_local(key: str, default_value: Union[None, str]=None):
    _assert_valid_key(key)
    kachery_cloud_dir = get_kachery_cloud_dir()
    file_name = f'{kachery_cloud_dir}/mutables/{key}'
    if not os.path.exists(file_name):
        return default_value
    with open(file_name, 'r') as f:
        return f.read()

def delete_mutable_local(key: str):
    kachery_cloud_dir = get_kachery_cloud_dir()
    a = key.split('/')
    last_str = a[-1]
    if last_str.startswith('@'):
        raise Exception(f'Mutable key is a folder: {key}')
    else:
        _assert_valid_key(key)
        file_name = f'{kachery_cloud_dir}/mutables/{key}'
        if not os.path.exists(file_name):
            raise Exception(f'Mutable does not exist: {key}')
        os.unlink(file_name)

def delete_mutable_folder_local(key: str):
    kachery_cloud_dir = get_kachery_cloud_dir()
    a = key.split('/')
    last_str = a[-1]
    if last_str.startswith('@'):
        # deleting a folder
        _assert_valid_folder_key(key)
        folder_name = f'{kachery_cloud_dir}/mutables/{key}'
        if not os.path.exists(folder_name):
            raise Exception(f'Mutable folder does not exist: {key}')
        shutil.rmtree(folder_name)
    else:
        raise Exception(f'Not a mutable folder: {key}')

# def _add_snails(x: str):
#     # some/path/to/key -> @some/@path/@to/key
#     a = x.split('/')
#     for i in range(len(a) - 1):
#         a[i] = '@' + a[i]
#     return '/'.join(a)


from email.policy import default
import os
import click
import kachery_cloud as kc


@click.group(help="Kachery cloud command-line client")
def cli():
    pass

@click.command(help="One-time initialization of the kachery cloud client")
def init():
    kc.init()

@click.command(help="Store file in kachery cloud")
@click.argument('filename')
@click.option('--cache-locally', is_flag=True)
@click.option('--label', required=False, default='')
def store_file(filename: str, cache_locally: bool, label: str):
    if label == '': label = os.path.basename(filename)
    uri = kc.store_file(filename, label=label, cache_locally=cache_locally)
    print(uri)

@click.command(help="Store file in kachery cloud")
@click.argument('filename')
@click.option('--label', required=False, default='')
def link_file(filename: str, label: str):
    if label == '': label = os.path.basename(filename)
    uri = kc.link_file(filename, label=label)
    print(uri)

@click.command(help="Load file from kachery cloud")
@click.argument('uri')
@click.option('--dest', required=False, default='')
def load_file(uri: str, dest: str):
    fname = kc.load_file(uri, dest=dest if len(dest) > 0 else None)
    if fname is not None:
        print(fname)

@click.command(help="Load file from kachery cloud")
@click.argument('uri')
def load_file_info(uri: str):
    import json
    x = kc.load_file_info(uri)
    if x is None:
        return
    print(json.dumps(x, indent=4))

@click.command(help="Load file from kachery cloud and write to stdout")
@click.argument('uri')
def cat_file(uri: str):
    kc.cat_file(uri)

@click.command(help="Store file in the local cache")
@click.argument('filename')
@click.option('--reference', is_flag=True)
@click.option('--label', required=False, default='')
def store_file_local(filename: str, reference: bool, label: str):
    if label == '': label = os.path.basename(filename)
    uri = kc.store_file_local(filename, label=label, reference=reference)
    print(uri)

@click.command(help="Share local files")
@click.option('--project', required=False, default='')
def share_local_files_experimental(project: str):
    project = project if project != '' else None
    kc.share_local_files_experimental(project_id=project)

@click.command(help="Request file")
@click.argument('uri')
@click.option('--project', required=True)
def request_file_experimental(uri: str, project: str):
    fname = kc.request_file_experimental(uri, project_id=project)
    print(fname)

cli.add_command(init)
cli.add_command(store_file)
cli.add_command(link_file)
cli.add_command(load_file)
cli.add_command(load_file_info)
cli.add_command(cat_file)
cli.add_command(store_file_local)
cli.add_command(share_local_files_experimental)
cli.add_command(request_file_experimental)


import os
from .init import _get_client_info

def get_project_id():
    project_id = os.getenv('KACHERY_CLOUD_PROJECT', '')
    if project_id:
        return project_id
    client_info = _get_client_info()
    if client_info is None:
        return None
    client = client_info['client']
    defaultProjectId = client.get('defaultProjectId', None)
    return defaultProjectId


import os
from typing import Union
import requests
import simplejson
import kachery_cloud as kcl

from ..get_project_id import get_project_id
from .._serialize import _deserialize
from .._json_stringify_deterministic import _json_stringify_deterministic
from .._get_project_bucket_base_url import _get_project_bucket_base_url
from .._kacherycloud_request import _kacherycloud_request

def upload_task_result(*, task_name: str, task_job_id: dict, serialized_task_result: dict, project_id: Union[str, None]=None):
    if project_id is None:
        project_id = get_project_id()
    task_result_text = _json_stringify_deterministic(serialized_task_result)
    size = len(task_result_text)

    if not project_id.startswith('local:'):
        payload = {
            'type': 'initiateTaskResultUpload',
            'taskName': task_name,
            'taskJobId': task_job_id,
            'size': size
        }
        payload['projectId'] = project_id
        response = _kacherycloud_request(payload)
        signed_upload_url = response['signedUploadUrl']
        resp_upload = requests.put(signed_upload_url, data=task_result_text)
        if resp_upload.status_code != 200:
            raise Exception(f'Error uploading task_result to bucket ({resp_upload.status_code}) {resp_upload.reason}: {resp_upload.text}')
        payload2 = {
            'type': 'finalizeTaskResultUpload',
            'taskName': task_name,
            'taskJobId': task_job_id,
            'size': size
        }
        payload2['projectId'] = project_id
        response2 = _kacherycloud_request(payload2)
    else:
        _store_task_result_text_locally(task_name=task_name, task_job_id=task_job_id, task_result_text=task_result_text)

def download_task_result(*, task_name: str, task_job_id: dict, project_id: str):
    if not project_id.startswith('local:'):
        project_bucket_base_url = _get_project_bucket_base_url(project_id)
        s = task_job_id
        url = f'{project_bucket_base_url}/taskResults/{task_name}/{s[0]}{s[1]}/{s[2]}{s[3]}/{s[4]}{s[5]}/{s}'
        resp = requests.get(url)
        if resp.status_code != 200:
            raise Exception(f'Error downloading task result ({resp.status_code}) {resp.reason}: {resp.text}')
        task_result_text = resp.content
    else:
        task_result_text = _load_task_result_text_locally(task_name=task_name, task_job_id=task_job_id)
    simplejson.loads(task_result_text)
    task_result_deserialized = _deserialize(task_result_text)
    return task_result_deserialized

def _store_task_result_text_locally(*, task_name: str, task_job_id: str, task_result_text: str):
    kcdir = kcl.get_kachery_cloud_dir()
    s = task_job_id
    parentdir = f'{kcdir}/task_results/{task_name}/{s[0]}{s[1]}/{s[2]}{s[3]}/{s[4]}{s[5]}'
    if not os.path.exists(parentdir):
        os.makedirs(parentdir)
    fname = f'{parentdir}/{s}'
    with open(fname, 'w') as f:
        f.write(task_result_text)

def _load_task_result_text_locally(*, task_name: str, task_job_id: str):
    kcdir = kcl.get_kachery_cloud_dir()
    s = task_job_id
    parentdir = f'{kcdir}/task_results/{task_name}/{s[0]}{s[1]}/{s[2]}{s[3]}/{s[4]}{s[5]}'
    fname = f'{parentdir}/{s}'
    with open(fname, 'r') as f:
        return f.read()


import os
import sys
from typing import Union
from .load_file import load_file


# returns None if write_to_stdout is True
def load_bytes(uri: str, start: Union[int, None], end: Union[int, None], *, write_to_stdout=False) -> bytes: 
    local_path = load_file(uri)
    if local_path is None:
        return None
    
    return _load_bytes_from_local_file(local_path, start=start, end=end, write_to_stdout=write_to_stdout)

def _load_bytes_from_local_file(local_fname: str, *, start: Union[int, None]=None, end: Union[int, None]=None, write_to_stdout: bool=False) -> bytes:
    size0 = os.path.getsize(local_fname)
    if start is None:
        start = 0
    if end is None:
        end = size0
    if start < 0 or start > size0 or end < start or end > size0:
        raise Exception('Invalid start/end range for file of size {}: {} - {}'.format(size0, start, end))
    if start == end:
        return bytes()
    with open(local_fname, 'rb') as f:
        f.seek(start)
        if write_to_stdout:
            ii = start
            while ii < end:
                nn = min(end - ii, 4096)
                data0 = f.read(nn)
                ii = ii + nn
                sys.stdout.buffer.write(data0)
            return None
        else:
            return f.read(end-start)


from typing import Union
from .TaskBackend.TaskClient import TaskClient, TaskErrorException
from .load_file import load_file
from ._access_group_encrypt import _access_group_decrypt


def request_file_experimental(uri: str, *, project_id: str, dest: Union[None, str]=None):
    # important to encrypt first before requesting the file
    if uri.startswith('sha1-enc://'):
        uri = _access_group_decrypt(uri)
    
    fname = load_file(uri, dest=dest)
    if fname is not None:
        return fname
    task_client = TaskClient(project_id=project_id)
    try:
        task_client.request_task(
            task_type='action',
            task_name='kachery_store_shared_file.1',
            task_input={
                'uri': uri
            }
        )
    except TaskErrorException as err:
        print(f'Error requesting file: {err.error_message}')
        return None
    return load_file(uri, dest=dest)



from typing import Any, Union, Protocol
import random
import math
import time

from .LocalPubsubListener import LocalPubsubListener

from ..get_project_id import get_project_id
from .upload_task_result import download_task_result
from .PubsubListener import PubsubListener
from .._kacherycloud_request import _kacherycloud_request
from ..get_client_id import get_client_id
from .._json_stringify_deterministic import _json_stringify_deterministic
from .._sha1_of_string import _sha1_of_string

class TaskRequestCallback(Protocol):
    def __call__(self, *, task_type: str, task_name: str, task_input: dict, task_job_id: str): ...

class TaskErrorException(Exception):
    def __init__(self, message: str) -> None:
        self.error_message = message
        super().__init__(message)

class TaskClient:
    def __init__(self, *, project_id: Union[None, str]=None, backend_id: Union[str, None]=None) -> None:
        if project_id is None:
            project_id = get_project_id()
        self._project_id = project_id
        self._backend_id = backend_id
    def request_task(self, *, task_type: str, task_name: str, task_input: dict):
        status: Union[str, None] = None
        result: Union[Any, None] = None
        error: Union[str, None] = None

        def handle_message(*, channel: str, message: dict):
            nonlocal status
            nonlocal error
            nonlocal result
            if message['type'] == 'setTaskStatus':
                if message['taskName'] == task_name:
                    if message['taskJobId'] == task_job_id:
                        status = message['status']
                        if status == 'error':
                            error = message['errorMessage']
                        elif status == 'finished':
                            if task_type == 'calculation':
                                result = download_task_result(task_name=task_name, task_job_id=task_job_id, project_id=self._project_id)
        
        if task_type == 'calculation':
            task_job_id = _sha1_of_string(_json_stringify_deterministic({
                'taskName': task_name,
                'taskInput': task_input
            }))
        else:
            task_job_id = _sha1_of_string(_random_string(100))
        
        if task_type == 'calculation':
            try:
                result = download_task_result(task_name=task_name, task_job_id=task_job_id, project_id=self._project_id)
            except:
                result = None
            if result is not None:
                return result
        
        request_task_message = {
            'type': 'requestTask',
            'taskType': task_type,
            'taskName': task_name,
            'taskInput': task_input,
            'taskJobId': task_job_id
        }
        if self._backend_id is not None:
            request_task_message['backendId'] = self._backend_id
        if not self._project_id.startswith('local:'):
            client_id = get_client_id()

            # subscribe to the provideTasks channel
            # to get notified when a task status has changed
            payload_sub = {
                'type': 'subscribeToPubsubChannel',
                'channelName': 'provideTasks'
            }
            payload_sub['projectId'] = self._project_id
            response_sub = _kacherycloud_request(payload_sub)
            subscribe_key = response_sub['subscribeKey']
            subscribe_token = response_sub['token']
            pubsub_channel_name = response_sub['pubsubChannelName']

            def renew_access_token():
                raise Exception('not used')
            listener = PubsubListener(
                channels=[pubsub_channel_name],
                uuid=client_id,
                subscribe_key=subscribe_key,
                access_token=subscribe_token,
                renew_access_token_callback=renew_access_token
            )
            listener.on_message(handle_message)
            listener.start()
            
            # request a task on the requestTasks pubsub channel
            payload = {
                'type': 'publishToPubsubChannel',
                'channelName': 'requestTasks',
                'message': request_task_message
            }
            if self._project_id is not None:
                payload['projectId'] = self._project_id
            else:
                payload['projectId'] = get_project_id()
            _kacherycloud_request(payload)
        else:
            from ._run_task import _get_port_from_project_id, _publish_local_pubsub_message
            port = _get_port_from_project_id(self._project_id)
            listener = LocalPubsubListener(channel=f'provideTasks', port=port)
            listener.on_message(handle_message)
            listener.start()
            _publish_local_pubsub_message(channel=f'requestTasks', message=request_task_message, port=port)

        announced_started = False
        try:
            while True:
                listener.wait(0.1)
                if status == 'finished':
                    return result
                elif status == 'error':
                    raise TaskErrorException(error)
                elif status == 'started':
                    if not announced_started:
                        print('Task started')
                        announced_started = True
        finally:
            listener.stop()

    def listen_for_task_requests(self, callback: TaskRequestCallback):
        def handle_message(*, channel: str, message: dict):
            if message['type'] == 'requestTask':
                task_type = message['taskType']
                task_name = message['taskName']
                task_input = message['taskInput']
                task_job_id = message['taskJobId']
                backend_id = message.get('backendId', None)
                if task_type == 'calculation':
                    check_task_job_id = _sha1_of_string(_json_stringify_deterministic({
                        'taskName': task_name,
                        'taskInput': task_input
                    }))
                    if check_task_job_id != task_job_id:
                        raise Exception('Mismatch between task job ID and computed task job ID')
                callback(task_type=task_type, task_name=task_name, task_input=task_input, task_job_id=task_job_id, backend_id=backend_id)
        if not self._project_id.startswith('local:'):
            def subscribe_to_request_tasks_channel():
                # subscribe to the requestTasks channel
                # to listen for incoming tasks
                payload = {
                    'type': 'subscribeToPubsubChannel',
                    'channelName': 'requestTasks'
                }
                if self._project_id is not None:
                    payload['projectId'] = self._project_id
                else:
                    payload['projectId'] = get_project_id()
                return _kacherycloud_request(payload)
            def renew_access_token():
                while True:
                    try:
                        r = subscribe_to_request_tasks_channel()
                        break
                    except Exception as e:
                        print(e)
                        print('Problem renewing access token. Will retry in 30 seconds...')
                        time.sleep(30)
                return r['token']
                    
            client_id = get_client_id()
            response = subscribe_to_request_tasks_channel()
            listener = PubsubListener(
                channels=[response['pubsubChannelName']],
                uuid=client_id,
                subscribe_key=response['subscribeKey'],
                access_token=response['token'],
                renew_access_token_callback=renew_access_token
            )
            listener.on_message(handle_message)
            listener.start()
            return listener
        else:
            from ._run_task import _get_port_from_project_id
            port = _get_port_from_project_id(self._project_id)
            listener = LocalPubsubListener(channel=f'requestTasks', port=port)
            listener.on_message(handle_message)
            listener.start()
            return listener

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))


from typing import Any
from numpy import floor
import simplejson


def _json_stringify_deterministic(x: Any):
    separators=(',', ':')
    x = _replace_float_by_int_when_appropriate(x)
    return simplejson.dumps(x, separators=separators, indent=None, allow_nan=False, sort_keys=True)

def _replace_float_by_int_when_appropriate(x):
    if isinstance(x, float):
        if x == floor(x):
            return int(x)
        else:
            return x
    elif isinstance(x, dict):
        ret = {}
        for k in x.keys():
            ret[k] = _replace_float_by_int_when_appropriate(x[k])
        return ret
    elif isinstance(x, tuple):
        return tuple([_replace_float_by_int_when_appropriate(a) for a in x])
    elif isinstance(x, list):
        return [_replace_float_by_int_when_appropriate(a) for a in x]
    else:
        return x


import os

_kachery_gateway_url_by_zone = {}

_default_kachery_gateway_url = 'https://kachery-gateway.figurl.org'

def _get_kachery_gateway_url():
    url = os.environ.get('KACHERY_GATEWAY_URL', None)
    if url is not None:
        return url
    kachery_zone = os.environ.get('KACHERY_ZONE', None)
    if kachery_zone is not None:
        if kachery_zone in _kachery_gateway_url_by_zone:
            return _kachery_gateway_url_by_zone[kachery_zone]
        from ._kachery_gateway_request import _kachery_gateway_request
        resp = _kachery_gateway_request({
            'type': 'getZoneInfo',
            'zoneName': kachery_zone
        }, kachery_gateway_url=_default_kachery_gateway_url) # important to use default here so we don't get an infinite recursion
        if resp['found']:
            url = resp['kacheryGatewayUrl']
            _kachery_gateway_url_by_zone[kachery_zone] = url
            return url
        else:
            raise Exception(f'Unrecognized kachery zone: {kachery_zone}')
    return _default_kachery_gateway_url